@@ -1,19 +1,18 @@
 #![cfg(feature = "full")]
 
 use {
-    crate::instruction::Instruction,
+    crate::{
+        feature_set::{
+            libsecp256k1_0_5_upgrade_enabled, libsecp256k1_fail_on_bad_count, FeatureSet,
+        },
+        instruction::Instruction,
+        precompiles::PrecompileError,
+    },
     digest::Digest,
     serde_derive::{Deserialize, Serialize},
+    std::sync::Arc,
 };
 
-#[derive(Debug, PartialEq)]
-pub enum Secp256k1Error {
-    InvalidSignature,
-    InvalidRecoveryId,
-    InvalidDataOffsets,
-    InvalidInstructionDataSize,
-}
-
 pub const HASHED_PUBKEY_SERIALIZED_SIZE: usize = 20;
 pub const SIGNATURE_SERIALIZED_SIZE: usize = 64;
 pub const SIGNATURE_OFFSETS_SERIALIZED_SIZE: usize = 11;
@@ -40,7 +39,7 @@ pub fn new_secp256k1_instruction(
     hasher.update(&message_arr);
     let message_hash = hasher.finalize();
     let mut message_hash_arr = [0u8; 32];
-    message_hash_arr.copy_from_slice(&message_hash.as_slice());
+    message_hash_arr.copy_from_slice(message_hash.as_slice());
     let message = libsecp256k1::Message::parse(&message_hash_arr);
     let (signature, recovery_id) = libsecp256k1::sign(&message, priv_key);
     let signature_arr = signature.serialize();
@@ -101,20 +100,27 @@ pub fn construct_eth_pubkey(
     addr
 }
 
-pub fn verify_eth_addresses(
+pub fn verify(
     data: &[u8],
     instruction_datas: &[&[u8]],
-    libsecp256k1_0_5_upgrade_enabled: bool,
-) -> Result<(), Secp256k1Error> {
+    feature_set: &Arc<FeatureSet>,
+) -> Result<(), PrecompileError> {
     if data.is_empty() {
-        return Err(Secp256k1Error::InvalidInstructionDataSize);
+        return Err(PrecompileError::InvalidInstructionDataSize);
     }
     let count = data[0] as usize;
+    if feature_set.is_active(&libsecp256k1_fail_on_bad_count::id()) && count == 0 && data.len() > 1
+    {
+        // count is zero but the instruction data indicates that is probably not
+        // correct, fail the instruction to catch probable invalid secp256k1
+        // instruction construction.
+        return Err(PrecompileError::InvalidInstructionDataSize);
+    }
     let expected_data_size = count
         .saturating_mul(SIGNATURE_OFFSETS_SERIALIZED_SIZE)
         .saturating_add(1);
     if data.len() < expected_data_size {
-        return Err(Secp256k1Error::InvalidInstructionDataSize);
+        return Err(PrecompileError::InvalidInstructionDataSize);
     }
     for i in 0..count {
         let start = i
@@ -123,21 +129,21 @@ pub fn verify_eth_addresses(
         let end = start.saturating_add(SIGNATURE_OFFSETS_SERIALIZED_SIZE);
 
         let offsets: SecpSignatureOffsets = bincode::deserialize(&data[start..end])
-            .map_err(|_| Secp256k1Error::InvalidSignature)?;
+            .map_err(|_| PrecompileError::InvalidSignature)?;
 
         // Parse out signature
         let signature_index = offsets.signature_instruction_index as usize;
         if signature_index >= instruction_datas.len() {
-            return Err(Secp256k1Error::InvalidInstructionDataSize);
+            return Err(PrecompileError::InvalidInstructionDataSize);
         }
         let signature_instruction = instruction_datas[signature_index];
         let sig_start = offsets.signature_offset as usize;
         let sig_end = sig_start.saturating_add(SIGNATURE_SERIALIZED_SIZE);
         if sig_end >= signature_instruction.len() {
-            return Err(Secp256k1Error::InvalidSignature);
+            return Err(PrecompileError::InvalidSignature);
         }
 
-        let sig_parse_result = if libsecp256k1_0_5_upgrade_enabled {
+        let sig_parse_result = if feature_set.is_active(&libsecp256k1_0_5_upgrade_enabled::id()) {
             libsecp256k1::Signature::parse_standard_slice(
                 &signature_instruction[sig_start..sig_end],
             )
@@ -147,10 +153,10 @@ pub fn verify_eth_addresses(
             )
         };
 
-        let signature = sig_parse_result.map_err(|_| Secp256k1Error::InvalidSignature)?;
+        let signature = sig_parse_result.map_err(|_| PrecompileError::InvalidSignature)?;
 
         let recovery_id = libsecp256k1::RecoveryId::parse(signature_instruction[sig_end])
-            .map_err(|_| Secp256k1Error::InvalidRecoveryId)?;
+            .map_err(|_| PrecompileError::InvalidRecoveryId)?;
 
         // Parse out pubkey
         let eth_address_slice = get_data_slice(
@@ -177,11 +183,11 @@ pub fn verify_eth_addresses(
             &signature,
             &recovery_id,
         )
-        .map_err(|_| Secp256k1Error::InvalidSignature)?;
+        .map_err(|_| PrecompileError::InvalidSignature)?;
         let eth_address = construct_eth_pubkey(&pubkey);
 
         if eth_address_slice != eth_address {
-            return Err(Secp256k1Error::InvalidSignature);
+            return Err(PrecompileError::InvalidSignature);
         }
     }
     Ok(())
@@ -192,16 +198,16 @@ fn get_data_slice<'a>(
     instruction_index: u8,
     offset_start: u16,
     size: usize,
-) -> Result<&'a [u8], Secp256k1Error> {
+) -> Result<&'a [u8], PrecompileError> {
     let signature_index = instruction_index as usize;
     if signature_index >= instruction_datas.len() {
-        return Err(Secp256k1Error::InvalidDataOffsets);
+        return Err(PrecompileError::InvalidDataOffsets);
     }
     let signature_instruction = &instruction_datas[signature_index];
     let start = offset_start as usize;
     let end = start.saturating_add(size);
     if end > signature_instruction.len() {
-        return Err(Secp256k1Error::InvalidSignature);
+        return Err(PrecompileError::InvalidSignature);
     }
 
     Ok(&instruction_datas[signature_index][start..end])
@@ -209,15 +215,38 @@ fn get_data_slice<'a>(
 
 #[cfg(test)]
 pub mod test {
-    use super::*;
+    use {
+        super::*,
+        crate::{
+            feature_set,
+            hash::Hash,
+            secp256k1_instruction::{
+                new_secp256k1_instruction, SecpSignatureOffsets, SIGNATURE_OFFSETS_SERIALIZED_SIZE,
+            },
+            signature::{Keypair, Signer},
+            transaction::Transaction,
+        },
+        rand::{thread_rng, Rng},
+        std::sync::Arc,
+    };
 
-    fn test_case(num_signatures: u8, offsets: &SecpSignatureOffsets) -> Result<(), Secp256k1Error> {
+    fn test_case(
+        num_signatures: u8,
+        offsets: &SecpSignatureOffsets,
+    ) -> Result<(), PrecompileError> {
         let mut instruction_data = vec![0u8; DATA_START];
         instruction_data[0] = num_signatures;
         let writer = std::io::Cursor::new(&mut instruction_data[1..]);
         bincode::serialize_into(writer, &offsets).unwrap();
+        let mut feature_set = FeatureSet::all_enabled();
+        feature_set
+            .active
+            .remove(&libsecp256k1_0_5_upgrade_enabled::id());
+        feature_set
+            .inactive
+            .insert(libsecp256k1_0_5_upgrade_enabled::id());
 
-        verify_eth_addresses(&instruction_data, &[&[0u8; 100]], false)
+        verify(&instruction_data, &[&[0u8; 100]], &Arc::new(feature_set))
     }
 
     #[test]
@@ -230,10 +259,17 @@ pub mod test {
         let writer = std::io::Cursor::new(&mut instruction_data[1..]);
         bincode::serialize_into(writer, &offsets).unwrap();
         instruction_data.truncate(instruction_data.len() - 1);
+        let mut feature_set = FeatureSet::all_enabled();
+        feature_set
+            .active
+            .remove(&libsecp256k1_0_5_upgrade_enabled::id());
+        feature_set
+            .inactive
+            .insert(libsecp256k1_0_5_upgrade_enabled::id());
 
         assert_eq!(
-            verify_eth_addresses(&instruction_data, &[&[0u8; 100]], false),
-            Err(Secp256k1Error::InvalidInstructionDataSize)
+            verify(&instruction_data, &[&[0u8; 100]], &Arc::new(feature_set)),
+            Err(PrecompileError::InvalidInstructionDataSize)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -242,7 +278,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidInstructionDataSize)
+            Err(PrecompileError::InvalidInstructionDataSize)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -251,7 +287,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidDataOffsets)
+            Err(PrecompileError::InvalidDataOffsets)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -260,7 +296,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidDataOffsets)
+            Err(PrecompileError::InvalidDataOffsets)
         );
     }
 
@@ -273,7 +309,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -283,7 +319,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -293,7 +329,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -303,7 +339,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
     }
 
@@ -315,7 +351,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -324,7 +360,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
     }
 
@@ -336,7 +372,7 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
 
         let offsets = SecpSignatureOffsets {
@@ -345,7 +381,72 @@ pub mod test {
         };
         assert_eq!(
             test_case(1, &offsets),
-            Err(Secp256k1Error::InvalidSignature)
+            Err(PrecompileError::InvalidSignature)
         );
     }
+
+    #[test]
+    fn test_count_is_zero_but_sig_data_exists() {
+        solana_logger::setup();
+
+        let mut instruction_data = vec![0u8; DATA_START];
+        let offsets = SecpSignatureOffsets::default();
+        instruction_data[0] = 0;
+        let writer = std::io::Cursor::new(&mut instruction_data[1..]);
+        bincode::serialize_into(writer, &offsets).unwrap();
+        let mut feature_set = FeatureSet::all_enabled();
+        feature_set
+            .active
+            .remove(&libsecp256k1_0_5_upgrade_enabled::id());
+        feature_set
+            .inactive
+            .insert(libsecp256k1_0_5_upgrade_enabled::id());
+
+        assert_eq!(
+            verify(&instruction_data, &[&[0u8; 100]], &Arc::new(feature_set)),
+            Err(PrecompileError::InvalidInstructionDataSize)
+        );
+    }
+
+    #[test]
+    fn test_secp256k1() {
+        solana_logger::setup();
+        let offsets = SecpSignatureOffsets::default();
+        assert_eq!(
+            bincode::serialized_size(&offsets).unwrap() as usize,
+            SIGNATURE_OFFSETS_SERIALIZED_SIZE
+        );
+
+        let secp_privkey = libsecp256k1::SecretKey::random(&mut thread_rng());
+        let message_arr = b"hello";
+        let mut secp_instruction = new_secp256k1_instruction(&secp_privkey, message_arr);
+        let mint_keypair = Keypair::new();
+        let mut feature_set = feature_set::FeatureSet::all_enabled();
+        feature_set
+            .active
+            .remove(&feature_set::libsecp256k1_0_5_upgrade_enabled::id());
+        feature_set
+            .inactive
+            .insert(feature_set::libsecp256k1_0_5_upgrade_enabled::id());
+        let feature_set = Arc::new(feature_set);
+
+        let tx = Transaction::new_signed_with_payer(
+            &[secp_instruction.clone()],
+            Some(&mint_keypair.pubkey()),
+            &[&mint_keypair],
+            Hash::default(),
+        );
+
+        assert!(tx.verify_precompiles(&feature_set).is_ok());
+
+        let index = thread_rng().gen_range(0, secp_instruction.data.len());
+        secp_instruction.data[index] = secp_instruction.data[index].wrapping_add(12);
+        let tx = Transaction::new_signed_with_payer(
+            &[secp_instruction],
+            Some(&mint_keypair.pubkey()),
+            &[&mint_keypair],
+            Hash::default(),
+        );
+        assert!(tx.verify_precompiles(&feature_set).is_err());
+    }
 }
