@@ -65,10 +65,11 @@ impl SanitizedTransaction {
         let signatures = tx.signatures;
         let message = match tx.message {
             VersionedMessage::Legacy(message) => SanitizedMessage::Legacy(message),
-            VersionedMessage::V0(message) => SanitizedMessage::V0(v0::LoadedMessage {
-                loaded_addresses: address_loader.load_addresses(&message.address_table_lookups)?,
-                message,
-            }),
+            VersionedMessage::V0(message) => {
+                let loaded_addresses =
+                    address_loader.load_addresses(&message.address_table_lookups)?;
+                SanitizedMessage::V0(v0::LoadedMessage::new(message, loaded_addresses))
+            }
         };
 
         let is_simple_vote_tx = is_simple_vote_tx.unwrap_or_else(|| {
@@ -139,7 +140,7 @@ impl SanitizedTransaction {
         match &self.message {
             SanitizedMessage::V0(sanitized_msg) => VersionedTransaction {
                 signatures,
-                message: VersionedMessage::V0(sanitized_msg.message.clone()),
+                message: VersionedMessage::V0(v0::Message::clone(&sanitized_msg.message)),
             },
             SanitizedMessage::Legacy(message) => VersionedTransaction {
                 signatures,
@@ -191,7 +192,7 @@ impl SanitizedTransaction {
     pub fn get_loaded_addresses(&self) -> LoadedAddresses {
         match &self.message {
             SanitizedMessage::Legacy(_) => LoadedAddresses::default(),
-            SanitizedMessage::V0(message) => message.loaded_addresses.clone(),
+            SanitizedMessage::V0(message) => LoadedAddresses::clone(&message.loaded_addresses),
         }
     }
 
@@ -204,7 +205,7 @@ impl SanitizedTransaction {
     fn message_data(&self) -> Vec<u8> {
         match &self.message {
             SanitizedMessage::Legacy(message) => message.serialize(),
-            SanitizedMessage::V0(message) => message.serialize(),
+            SanitizedMessage::V0(loaded_msg) => loaded_msg.message.serialize(),
         }
     }
 
